/**
 * Copyright (c) 2023 mol* contributors, licensed under MIT, See LICENSE file for more info.
 *
 * @author Adam Midlik <midlik@gmail.com>
 */
import { __assign, __spreadArray } from "tslib";
import { onelinerJsonString } from '../../../../mol-util/json';
import { isPlainObject, mapObjectMap } from '../../../../mol-util/object';
import { AllRequired, paramsValidationIssues } from './params-schema';
import { treeToString } from './tree-utils';
/** Get params from a tree node */
export function getParams(node) {
    var _a;
    return (_a = node.params) !== null && _a !== void 0 ? _a : {};
}
/** Get children from a tree node */
export function getChildren(tree) {
    var _a;
    return (_a = tree.children) !== null && _a !== void 0 ? _a : [];
}
export function TreeSchema(schema) {
    return schema;
}
export function TreeSchemaWithAllRequired(schema) {
    return __assign(__assign({}, schema), { nodes: mapObjectMap(schema.nodes, function (node) { return (__assign(__assign({}, node), { params: AllRequired(node.params) })); }) });
}
/** Return `undefined` if a tree conforms to the given schema,
 * return validation issues (as a list of lines) if it does not conform.
 * If `options.requireAll`, all parameters (including optional) must have a value provided.
 * If `options.noExtra` is true, presence of any extra parameters is treated as an issue.
 * If `options.anyRoot` is true, the kind of the root node is not enforced.
 */
export function treeValidationIssues(schema, tree, options) {
    if (options === void 0) { options = {}; }
    if (!isPlainObject(tree))
        return ["Node must be an object, not ".concat(tree)];
    if (!options.anyRoot && tree.kind !== schema.rootKind)
        return ["Invalid root node kind \"".concat(tree.kind, "\", root must be of kind \"").concat(schema.rootKind, "\"")];
    var nodeSchema = schema.nodes[tree.kind];
    if (!nodeSchema)
        return ["Unknown node kind \"".concat(tree.kind, "\"")];
    if (nodeSchema.parent && (options.parent !== undefined) && !nodeSchema.parent.includes(options.parent)) {
        return ["Node of kind \"".concat(tree.kind, "\" cannot appear as a child of \"").concat(options.parent, "\". Allowed parents for \"").concat(tree.kind, "\" are: ").concat(nodeSchema.parent.map(function (s) { return "\"".concat(s, "\""); }).join(', '))];
    }
    var issues = paramsValidationIssues(nodeSchema.params, getParams(tree), options);
    if (issues)
        return __spreadArray(["Invalid parameters for node of kind \"".concat(tree.kind, "\":")], issues.map(function (s) { return '  ' + s; }), true);
    for (var _i = 0, _a = getChildren(tree); _i < _a.length; _i++) {
        var child = _a[_i];
        var issues_1 = treeValidationIssues(schema, child, __assign(__assign({}, options), { anyRoot: true, parent: tree.kind }));
        if (issues_1)
            return issues_1;
    }
    return undefined;
}
/** Validate a tree against the given schema.
 * Do nothing if OK; print validation issues on console and throw an error is the tree does not conform.
 * Include `label` in the printed output. */
export function validateTree(schema, tree, label) {
    var issues = treeValidationIssues(schema, tree, { noExtra: true });
    if (issues) {
        console.warn("Invalid ".concat(label, " tree:\n").concat(treeToString(tree)));
        console.error("".concat(label, " tree validation issues:"));
        for (var _i = 0, issues_2 = issues; _i < issues_2.length; _i++) {
            var line = issues_2[_i];
            console.error(' ', line);
        }
        throw new Error('FormatError');
    }
}
/** Return documentation for a tree schema as plain text */
export function treeSchemaToString(schema, defaults) {
    return treeSchemaToString_(schema, defaults, false);
}
/** Return documentation for a tree schema as markdown text */
export function treeSchemaToMarkdown(schema, defaults) {
    return treeSchemaToString_(schema, defaults, true);
}
function treeSchemaToString_(schema, defaults, markdown) {
    var _a;
    if (markdown === void 0) { markdown = false; }
    var out = [];
    var bold = function (str) { return markdown ? "**".concat(str, "**") : str; };
    var code = function (str) { return markdown ? "`".concat(str, "`") : str; };
    var h1 = markdown ? '## ' : '  - ';
    var p1 = markdown ? '' : '    ';
    var h2 = markdown ? '- ' : '      - ';
    var p2 = markdown ? '  ' : '        ';
    var newline = markdown ? '\n\n' : '\n';
    out.push("Tree schema:");
    for (var kind in schema.nodes) {
        var _b = schema.nodes[kind], description = _b.description, params = _b.params, parent_1 = _b.parent;
        out.push("".concat(h1).concat(code(kind)));
        if (kind === schema.rootKind) {
            out.push("".concat(p1, "[Root of the tree must be of this kind]"));
        }
        if (description) {
            out.push("".concat(p1).concat(description));
        }
        out.push("".concat(p1, "Parent: ").concat(!parent_1 ? 'any' : parent_1.length === 0 ? 'none' : parent_1.map(code).join(' or ')));
        out.push("".concat(p1, "Params:").concat(Object.keys(params).length > 0 ? '' : ' none'));
        for (var key in params) {
            var field = params[key];
            var typeString = field.type.name;
            if (typeString.startsWith('(') && typeString.endsWith(')')) {
                typeString = typeString.slice(1, -1);
            }
            out.push("".concat(h2).concat(bold(code(key + (field.required ? ': ' : '?: ')))).concat(code(typeString)));
            var defaultValue = (_a = defaults === null || defaults === void 0 ? void 0 : defaults[kind]) === null || _a === void 0 ? void 0 : _a[key];
            if (field.description) {
                out.push("".concat(p2).concat(field.description));
            }
            if (defaultValue !== undefined) {
                out.push("".concat(p2, "Default: ").concat(code(onelinerJsonString(defaultValue))));
            }
        }
    }
    return out.join(newline);
}
